require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/finman/financialtransaction/transfer_create'
require 'uu/finman/financialtransaction/transfer_attributes'
require 'uu/finman/financialtransaction/transfer_set_attributes'

# Module Unicorn Universe
module UU
  # Module Finman
  module Finman
    # Module of finman Financial transaction
    module FinancialTransaction
      #Module of Transfer
      module Transfer
        extend self

        # Service path
        PATH = 'uu/finman/financialtransaction/FinancialTransaction'

        # Creates a new Transfer. The command creates a new Transfer for specified business case
        #
        # <b>Required attributes:</b><br>
        #   control_business_case_uri, counterparty_business_case_uri, name, amount_without_vat, vat_rate or amount_with_vat, financial_transaction_type,
        #   is_transaction_credit, realization_probability, expected_pay_date, transaction_currency, (credit_account_number and credit_side_currrency) or credit_account_uri,
        #   (debit_account_number and debit_side_currency) or debit_account_uri
        # Exactly one of attributes <code>credit_account_number</code> and <code>credit_account_uri</code> has to be set.<br>
        # Exactly one of attributes <code>debit_account_number</code> and <code>debit_account_uri</code> has to be set.<br>
        # Exactly one of attributes <code>vat_rate</code> and <code>amount_with_vat</code> has to be set. Both attributes can be set simultaneously only if following statement is true : <code>amount_without_vat</code> * <code>vat_rate</code> = <code>amount_with_rate</code>.<br>
        # Attribute <code>credit_side_currency</code> and <code>debit_side_currency</code> must be set if respective account is set by number.
        # Attribute <code>transaction_currency</code> must be set and must be one from <code>credit_side_currency</code> and <code>debit_side_currency</code>, with exception if both <code>credit_side_currency</code> and <code>debit_side_currency</code> are equal.
        # Attribute <code>related_links</code> must be valid JSON string with specific array structure with attributes UESURI of financial transaction and label. Both attributes are required. See example.
        #
        # @param [UU::OS::UESURI] control_business_case_uri UESURI of the business case of the control side to create Transfer for.
        # @param [UU::Finman::FinancialTransaction::TransferCreate] transfer DTO containing attributes of the new Transfer
        #
        # @return [UU::OS::UESURI] UESURI of the new Transfer. In case of setting <code>dry_run</code> parameter to <code>true</code>, <code>UU::OS::UESURI::NIL_URI</code> is returned instead.
        #
        #
        # @example
        #      # Prepare Business Case URI, which we want to use to create Transfer.
        #      control_business_case_uri = UU::OS::UESURI.new('ues:TERRITORY:BUSINESS_CASE');
        #
        #      # Create Transfer
        #        transfer_uri = UU::Finman::FinancialTransaction::Transfer.create(control_business_case_uri,
        #        :name => "Transfer",
        #        :credit_description => "kreditní popis",
        #        :debit_description => "debetní popis",
        #        :credit_financial_category => UU::OS::UESURI.new("ues:TERRITORY:APPS.UU.FINMAN/CONFIGURATION:SYS.FIN.CONFIG/TX_TYPE/CR.TXWER"),
        #        :debit_financial_category =>  UU::OS::UESURI.new("ues:TERRITORY:APPS.UU.FINMAN/CONFIGURATION:SYS.FIN.CONFIG/TX_TYPE/DR.TXOTR"),
        #        :realization_probability => "50",
        #        :is_transaction_credit => true,
        #        :financial_transaction_type => "NONCASH",
        #        :vat_rate => "10",
        #        :original_currency => "USD",
        #        :Vat_date => "2014-11-03",
        #        :expected_pay_date => "2014-11-06",
        #        :counterparty_business_case_uri => UU::OS::UESURI.new("ues:TERRITORY:CP_BC"),
        #        :credit_account_number => "kreditní číslo učtu",
        #        :debit_account_number => "debetní číslo učtu",
        #        :contract_uri => UU::OS::UESURI.new("ues:TERRITORY:CONTRACT"),
        #        :is_ready_to_be_paid => true,
        #        :credit_side_currency => "CZK",
        #        :debit_side_currency => "EUR",
        #        :transaction_currency => "EUR",
        #        :related_links => "[["ues:[TERRITORY_OID]:[FIN_TR_OID]:","label 1"],["ues:[TERRITORY_OID]:[FIN_TR_OID]:","label 2"]]"
        #       )
        def create(control_business_case_uri, transfer)
          svc = UU::OS::REST::RemoteClient.new(Transfer)
          transfer_dto = UU::Finman::FinancialTransaction::TransferCreate.new(transfer)
          # adjust given transfer dto to match expected format of financial transaction command
          transaction = transform_transfer_create_to_transaction_dto(transfer_dto)
          financial_transaction_create = transaction.to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('create', control_business_case_uri, financial_transaction_create)
            if res.nil?
              return res
            else
              return UU::OS::UESURI.new(res)
            end
          end
        end

        # Obtains Transfer attributes from service and returns them
        #
        # @param [String, UU::OS::UESURI] transfer_uri UESURI of the Transfer
        #
        # @return [TransferAttributes] Transfer attributes
        #
        #
        # @example
        #      # Prepare Transfer location which we want get attributes.
        #      transfer_uri = UU::OS::UESURI.new('ues:TERRITORY:TRANSFER');
        #
        #      # Get Transfer
        #      UU::Finman::FinancialTransaction::Transfer.get_attributes(transfer_uri);
        def get_attributes(transfer_uri)
          svc = UU::OS::REST::RemoteClient.new(Transfer)
          UU::OS::QoS::QoSHandler.auto_retry do
            attrs_string = svc.get('getAttributes', transfer_uri)
            attrs = JSON.parse(attrs_string, :symbolize_names => true)
            attrs = remove_invalid_keys(attrs)
            return UU::Finman::FinancialTransaction::TransferAttributes.new(attrs)
          end
        end
        # Sets attributes of Financial transfer.
        #
        # @param [String,UU::OS::UESURI] transfer_uri UESURI of modified Financial transfer
        # @param [TransferSetAttributes] transfer_set_attributes DTO containing modified attributes of the Financial transfer
        #
        # @return [UU::OS::UESURI] UESURI of modified transfer
        #
        #
        # @example
        #      # Prepare Transfer location which we want set attributes.
        #      transfer_uri = UU::OS::UESURI.new('ues:TERRITORY:TRANSFER');
        #
        #      # Set Transfer attributes
        #      UU::Finman::FinancialTransaction::Transfer.set_attributes(transfer_uri,
        #        :name => "Transfer set");
        def set_attributes(transfer_uri, transfer_set_attributes)
          svc = UU::OS::REST::RemoteClient.new(Transfer)
          transfer_dto = UU::Finman::FinancialTransaction::TransferSetAttributes.new(transfer_set_attributes)
          # adjust given transfer dto to match expected format of financial transaction command
          transaction = transform_modify_transfer_to_transaction_dto(transfer_dto)
          financial_transaction_dto = transaction.to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('setAttributes', transfer_uri, financial_transaction_dto)
            return UU::OS::UESURI.new(res)
          end
        end

        # Adjust dto of Transfer to match requirements of financial transaction dto of server side.
        # @private
        def transform_transfer_create_to_transaction_dto(transfer_dto)
          if transfer_dto.is_a?(UU::OS::REST::DTO)
            attributes = transfer_dto.to_hash
            # set class of transaction to determine Transfer at the server side.
            attributes[:financialTransactionClass] = "TRANSFER"
            #transform attribute to another by attributes of financial transaction
            attributes[:isReadyForAcceptance] = attributes[:isReadyToBePaid]
            attributes[:creditVatDate] = attributes[:vatDate]
            # remove unwanted attributes (in case that e.g. hashmap with attributes from other FinTx was modified and used)
            attributes.delete(:payDate)
            attributes.delete(:metaArtifactUri)
            attributes.delete(:requestApprovalByManager)
            attributes.delete(:waitForAcceptanceByAccountant)
            attributes.delete(:creditBusinessCaseFromOppositeTerritory)
            attributes.delete(:debitBusinessCaseFromOppositeTerritory)
            attributes.delete(:debitVatDate)
            attributes.delete(:dueDate)
            attributes.delete(:vatDate)
            attributes.delete(:forAcceptanceDate)
            attributes.delete(:invoiceNumber)
            attributes.delete(:isPlanned)
            attributes.delete(:isReadyToBePaid)
            attributes.delete(:protocolDate)
            attributes.delete(:constantSymbol)
            attributes.delete(:specificSymbol)
            return attributes
          end
          fail "Given object is neither UU::OS::REST::DTO nor its descendant!"
        end
        private :transform_transfer_create_to_transaction_dto

        # Hash returned from service contains all attributes from financial transaction.
        # Remove invalid keys to contain only values for this specific class of transaction.
        # @private
        def remove_invalid_keys(hash)
          if hash.is_a?(Hash)
            hash[:vatDate] = hash[:creditVatDate]
            hash[:isReadyToBePaid] = hash[:isReadyForAcceptance]
            hash.delete(:creditBusinessCaseFromOppositeTerritory)
            hash.delete(:debitBusinessCaseFromOppositeTerritory)
            hash.delete(:constantSymbol)
            hash.delete(:specificSymbol)
            hash.delete(:acceptedDate)
            hash.delete(:forAcceptanceDate)
            hash.delete(:creditVatDate)
            hash.delete(:debitVatDate)
            hash.delete(:dueDate)
            hash.delete(:invoiceNumber)
            hash.delete(:interterritorialPaymentCopyUri)
            hash.delete(:interterritorialPaymentOriginalUri)
            hash.delete(:isPlanned)
            hash.delete(:isReadyForAcceptance)
            hash.delete(:protocolDate)
            hash.delete(:requestApprovalByManager)
            hash.delete(:waitForAcceptanceByAccountant)
            hash.delete(:closeAfterCreation)
            hash.delete(:accountFromOppositeTerritory)
            return hash
          end
          fail "Given object must be Hash!"
        end
        private :remove_invalid_keys

        # Adjust dto of Transfer to match requirements of financial transaction dto of server side.
        # @private
        def transform_modify_transfer_to_transaction_dto(transfer_dto)
          if transfer_dto.is_a?(UU::OS::REST::DTO)
            attributes = transfer_dto.to_hash
            # set class of transaction to determine Transfer at the server side.
            attributes[:financialTransactionClass] = "TRANSFER"
            #transform attribute to another by attributes of financial transaction
            attributes[:creditVatDate] = attributes[:vatDate]
            # remove unwanted attributes (in case that e.g. hashmap with attributes from other FinTx was modified and used)
            attributes.delete(:payDate)
            attributes.delete(:requestApprovalByManager)
            attributes.delete(:waitForAcceptanceByAccountant)
            attributes.delete(:creditBusinessCaseFromOppositeTerritory)
            attributes.delete(:debitBusinessCaseFromOppositeTerritory)
            attributes.delete(:debitVatDate)
            attributes.delete(:dueDate)
            attributes.delete(:vatDate)
            attributes.delete(:forAcceptanceDate)
            attributes.delete(:invoiceNumber)
            attributes.delete(:isPlanned)
            attributes.delete(:protocolDate)
            attributes.delete(:constantSymbol)
            attributes.delete(:specificSymbol)
            return attributes
          end
          fail "Given object is neither UU::OS::REST::DTO nor its descendant!"
        end
        private :transform_modify_transfer_to_transaction_dto
      end
    end
  end
end
